;; Generated package description from consult-denote.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "consult-denote" "0.4.0.0.20251030.114409" "Use Consult in tandem with Denote" '((emacs "28.1") (denote "4.0.0") (consult "2.2")) :commit "f73cdb84392dd30dd8b3746f2020a47ce1541113" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :url "https://github.com/protesilaos/consult-denote")
